import logging
import sys
from importlib import metadata

logger = logging.getLogger(__package__)

# If the plugin is running in a TTY, we assume it is used through MLflow CLI
# so we output INFO logs to the console.
# Otherwise we're a well-behaved library and do not set logging configuration.
if sys.stdout.isatty():
    logger.setLevel(logging.INFO)

    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)

    logger.addHandler(handler)


__version__ = metadata.version(__package__)



"""Templated Knative manifests helper module."""
import logging
from importlib.resources import files
from pathlib import Path
from string import Template
from typing import Any

import yaml

DEFAULT_SERVICE_TEMPLATE = Template(
    files(__package__).joinpath("service.yaml").read_text()
)

logger = logging.getLogger(__package__)


def get_service_body(
    namespace: str, name: str, image: str, template_path: str | Path | None = None
) -> Any:
    """Substitute variables in a Knative service template and return the result.

    If no template file is provided, a default template is used.
    """
    template = DEFAULT_SERVICE_TEMPLATE
    if template_path is not None:
        logger.debug("Loading template from '%s' for service '%s'", template_path, name)

        with open(template_path) as file:
            template = Template(file.read())
    else:
        logger.debug("Using default template for service '%s'", name)

    return yaml.safe_load(
        template.substitute(
            namespace=namespace,
            name=name,
            image=image,
        )
    )



"""MLflow Knative Deployment Plugin.

The plugin implements a MLflow deployment client for Knative targets
as well as helper functions.

Implementing: https://mlflow.org/docs/latest/_modules/mlflow/deployments/base.html
"""
import logging
import warnings
from http import HTTPStatus

import docker
import kubernetes
import mlflow
import pandas
import requests

from .knative import KnativeServingV1Api, KnativeTimeoutError
from .mlflow_docker import push_model_as_image
from .templating import get_service_body

MLFLOW_MODEL_FLAVOR_PYFUNC = "python_function"
IMAGE_REPOSITORY_CONFIG_KEY = "image_repository"
IMAGE_TAG_CONFIG_KEY = "image_tag"
DEFAULT_IMAGE_TAG = "latest"

NAMESPACE_CONFIG_KEY = "namespace"
DEFAULT_NAMESPACE = "default"

TEMPLATE_PATH_CONFIG_KEY = "service_template"

logger = logging.getLogger(__package__)


class EndpointArgumentNotSupported(mlflow.exceptions.MlflowException):
    """Raised when the unsupported 'endpoint' argument is used."""

    def __init__(self):
        super().__init__(
            "'endpoint' argument not supported by Knative deployment plugin."
        )


class KnativeDeploymentClient(mlflow.deployments.BaseDeploymentClient):
    """Knative Deployment Client for MLflow."""

    def __init__(self, uri: str):
        super().__init__(target_uri=uri)
        self._docker_client = docker.from_env()
        self._kubernetes_client = kubernetes.config.new_client_from_config(
            context=uri.removeprefix("knative:/")
        )

    def create_deployment(
        self,
        name: str,
        model_uri: str,
        flavor: str | None = None,
        config: dict[str, str] | None = None,
        endpoint: str | None = None,
    ) -> dict[str, str]:
        """Deploy a model to a Knative target."""
        if endpoint:
            raise EndpointArgumentNotSupported

        flavor = flavor if flavor is not None else MLFLOW_MODEL_FLAVOR_PYFUNC
        if flavor != MLFLOW_MODEL_FLAVOR_PYFUNC:
            raise mlflow.exceptions.MlflowException(
                f"only '{MLFLOW_MODEL_FLAVOR_PYFUNC}' model flavor is supported."
            )

        logger.info(
            "Creating deployment '%s' from model '%s' targetting '%s'",
            name,
            model_uri,
            self.target_uri,
        )

        config = config if config is not None else {}
        tag = config.get(IMAGE_TAG_CONFIG_KEY, DEFAULT_IMAGE_TAG)
        image = push_model_as_image(
            self._docker_client, model_uri, config[IMAGE_REPOSITORY_CONFIG_KEY], tag
        )

        with self._kubernetes_client as kubernetes_client:
            knative_api = KnativeServingV1Api(kubernetes_client)
            namespace = config.get(NAMESPACE_CONFIG_KEY, DEFAULT_NAMESPACE)
            try:
                service = knative_api.create_namespaced_service(
                    namespace=namespace,
                    body=get_service_body(
                        namespace=namespace,
                        name=name,
                        image=image,
                        template_path=config.get(TEMPLATE_PATH_CONFIG_KEY),
                    ),
                )
            except kubernetes.client.rest.ApiException as error:
                if error.status == HTTPStatus.CONFLICT:
                    raise mlflow.exceptions.MlflowException(
                        f"a deployment with name '{name}' "
                        f"already exists in namespace '{namespace}'."
                    )
                else:
                    raise mlflow.exceptions.MlflowException(
                        f"Kubernetes API error ({error.status})"
                    ) from error
            except KnativeTimeoutError as error:
                raise mlflow.exceptions.MlflowException(
                    f"service creation failed: {str(error)}"
                )

        return {"name": service["metadata"]["name"], "flavor": flavor}

    def update_deployment(
        self,
        name: str,
        model_uri: str,
        flavor: str | None = None,
        config: dict[str, str] | None = None,
        endpoint: str | None = None,
    ) -> dict[str, str]:
        """Update a model on a Knative target."""
        if endpoint:
            raise EndpointArgumentNotSupported

        flavor = flavor if flavor is not None else MLFLOW_MODEL_FLAVOR_PYFUNC
        if flavor != MLFLOW_MODEL_FLAVOR_PYFUNC:
            raise mlflow.exceptions.MlflowException(
                f"only '{MLFLOW_MODEL_FLAVOR_PYFUNC}' model flavor is supported."
            )

        logger.info(
            "Updating deployment '%s' from model '%s' targetting '%s'",
            name,
            model_uri,
            self.target_uri,
        )

        config = config if config is not None else {}
        tag = config.get(IMAGE_TAG_CONFIG_KEY, DEFAULT_IMAGE_TAG)
        image = push_model_as_image(
            self._docker_client, model_uri, config[IMAGE_REPOSITORY_CONFIG_KEY], tag
        )

        with self._kubernetes_client as kubernetes_client:
            knative_api = KnativeServingV1Api(kubernetes_client)
            namespace = config.get(NAMESPACE_CONFIG_KEY, DEFAULT_NAMESPACE)
            try:
                service = knative_api.update_namespaced_service(
                    namespace=namespace,
                    name=name,
                    body=get_service_body(
                        namespace=namespace,
                        name=name,
                        image=image,
                        template_path=config.get(TEMPLATE_PATH_CONFIG_KEY),
                    ),
                )
            except kubernetes.client.rest.ApiException as error:
                if error.status == HTTPStatus.NOT_FOUND:
                    raise mlflow.exceptions.MlflowException(
                        f"no deployment with name '{name}' "
                        f"found in namespace '{namespace}'."
                    )
                else:
                    raise mlflow.exceptions.MlflowException(
                        f"Kubernetes API error ({error.status})"
                    ) from error
            except KnativeTimeoutError as error:
                raise mlflow.exceptions.MlflowException(
                    f"service update failed: {str(error)}"
                )

        return {"name": service["metadata"]["name"], "flavor": flavor}

    def delete_deployment(
        self,
        name: str,
        config: dict[str, str] | None = None,
        endpoint: str | None = None,
    ) -> None:
        """Delete a model on a Knative target."""
        if endpoint:
            raise EndpointArgumentNotSupported

        logger.info("Deleting deployment '%s' on target '%s'", name, self.target_uri)

        config = config if config is not None else {}

        with self._kubernetes_client as kubernetes_client:
            knative_api = KnativeServingV1Api(kubernetes_client)
            namespace = config.get(NAMESPACE_CONFIG_KEY, DEFAULT_NAMESPACE)
            try:
                knative_api.delete_namespaced_service(namespace=namespace, name=name)
            except kubernetes.client.rest.ApiException as error:
                if error.status == HTTPStatus.NOT_FOUND:
                    raise mlflow.exceptions.MlflowException(
                        f"no deployment with name '{name}' "
                        f"found in namespace '{namespace}'."
                    )
                else:
                    raise mlflow.exceptions.MlflowException(
                        f"Kubernetes API error ({error.status})"
                    ) from error

    def list_deployments(self, endpoint: str | None = None) -> list[dict[str, str]]:
        """List deployments on a Knative target (across all namespaces)."""
        if endpoint:
            raise EndpointArgumentNotSupported

        with self._kubernetes_client as kubernetes_client:
            core_api = kubernetes.client.CoreV1Api(kubernetes_client)
            knative_api = KnativeServingV1Api(kubernetes_client)

            return [
                {
                    "name": service["metadata"]["name"],
                    "namespace": service["metadata"]["namespace"],
                    "url": service["status"]["url"],
                }
                for namespace in core_api.list_namespace().items
                for service in knative_api.list_namespaced_service(
                    namespace=namespace.metadata.name
                )["items"]
            ]

    def get_deployment(self, name: str, endpoint: str | None = None) -> dict[str, str]:
        """Get metadata for a deployment on a Knative target (across all namespaces).

        Note: this fails if multiple deployments with the same name exist across
        multiple namespaces.
        Unfortunately MLflow deployment client interface does not allow passing
        a `config` dict with a specified namespace to this method.
        """
        if endpoint:
            raise EndpointArgumentNotSupported

        with self._kubernetes_client as kubernetes_client:
            core_api = kubernetes.client.CoreV1Api(kubernetes_client)
            knative_api = KnativeServingV1Api(kubernetes_client)

            deployments = []
            for namespace in core_api.list_namespace().items:
                try:
                    service = knative_api.get_namespaced_service(
                        namespace=namespace.metadata.name, name=name
                    )
                    deployments.append(
                        {
                            "name": service["metadata"]["name"],
                            "namespace": service["metadata"]["namespace"],
                            "url": service["status"]["url"],
                        }
                    )
                except kubernetes.client.rest.ApiException as error:
                    if error.status == HTTPStatus.NOT_FOUND:
                        pass

            try:
                (deployment,) = deployments
            except ValueError as error:
                raise mlflow.exceptions.MlflowException(
                    "must return exactly one deployment matching the provided name."
                ) from error

            return deployment

    def predict(
        self,
        deployment_name: str | None = None,
        inputs: pandas.DataFrame | None = None,
        endpoint: str | None = None,
    ) -> list[dict[str, str]]:
        """Compute predictions on inputs using the specified Knative deployment."""
        if not deployment_name:
            raise mlflow.exceptions.MlflowException("a deployment name is required.")

        if endpoint:
            raise EndpointArgumentNotSupported

        inputs = inputs if inputs is not None else pandas.DataFrame()

        deployment = self.get_deployment(deployment_name)
        try:
            response = requests.post(
                f"{deployment['url']}/invocations",
                headers={"Content-Type": "application/json"},
                data=f"{{\"dataframe_split\": {inputs.to_json(orient='split')}}}",
            )
            response.raise_for_status()
        except requests.exceptions.ConnectionError as error:
            raise mlflow.exceptions.MlflowException(
                "could not connect to the deployment, are you sure "
                f"the service URL '{deployment['url']}' is publicly available?"
            ) from error
        except requests.exceptions.HTTPError as error:
            raise mlflow.exceptions.MlflowException(
                "deployment request HTTP error."
            ) from error

        return response.json()


def run_local(
    name: str,
    model_uri: str,
    flavor: str | None = None,
    config: dict[str, str] | None = None,
) -> None:
    """Run a model locally and serve it as an HTTP ReST API.

    Internally, we simply use `model.serve()`.
    This is mostly useful to check the "pyfunc" works before packaging.
    """
    flavor = flavor if flavor is not None else MLFLOW_MODEL_FLAVOR_PYFUNC
    if flavor != MLFLOW_MODEL_FLAVOR_PYFUNC:
        raise mlflow.exceptions.MlflowException(
            f"only '{MLFLOW_MODEL_FLAVOR_PYFUNC}' model flavor is supported."
        )

    config = config if config is not None else {}

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        model = mlflow.pyfunc.load_model(model_uri)

    model.serve(enable_mlserver=True, host=config.get("host"), port=config.get("port"))


def target_help() -> str | None:
    """MLflow deployments Knative target.

    * A `target_uri` must be constructed with the scheme "knative:/<kube context>".
      (e.g. for AWS EKS clusters, the context is the ARN of the cluster
      https://docs.aws.amazon.com/eks/latest/userguide/connecting-cluster.html)

    * `create_deployment` and `update_deployment` support `config` fields:

      * `image_repository` which specifies the Docker repository to push images
        used by Knative services deployments.

      * `image_tag` which specifies the Docker image tag for the image.

      * `namespace` is the Kubernetes cluster namespace for the service.

      * `service_template` which allows providing a file path to a template YAML
        manifest for the Knative service to deploy.
        Templated values `name`, `namespace` and `image` are replaced.

    * When using `predict` make sure the Knative service is publicly available.
    """
    return target_help.__doc__



"""MLflow Docker helpers."""
import io
import json
import logging
import warnings
from contextlib import redirect_stderr, redirect_stdout

import docker
import mlflow

logger = logging.getLogger(__package__)


def _mlflow_build_docker_image(model_uri: str, image: str) -> None:
    logger.info("Building image '%s' from '%s'", image, model_uri)

    with warnings.catch_warnings(), redirect_stdout(io.StringIO()), redirect_stderr(
        io.StringIO()
    ) as output:
        warnings.simplefilter("ignore")
        mlflow.models.build_docker(
            model_uri=model_uri,
            name=image,
            enable_mlserver=True,
        )
        build_log = output.getvalue()
        if "ERROR" in build_log:
            raise mlflow.exceptions.MlflowException(build_log)


def _docker_safe_push_image(client, repository: str, tag: str | None = None) -> None:
    logger.info("Pushing image '%s:%s'", repository, tag)

    response = client.images.push(repository, tag=tag)
    data = json.loads(response.strip().split("\n").pop())
    if "error" in data and data["error"]:
        raise docker.errors.APIError(data["error"])


def push_model_as_image(
    client, model_uri: str, repository: str, tag: str | None = None
) -> str:
    """Build a Docker image from a MLFlow model, push it to a Docker repository."""
    image = f"{repository}:{tag}" if tag else repository
    _mlflow_build_docker_image(model_uri, image)

    try:
        _docker_safe_push_image(client, repository, tag)
    except docker.errors.APIError as error:
        raise mlflow.exceptions.MlflowException(
            f"failed to push image '{image}'."
        ) from error

    return image



"""Knative Serving API module with event handling helpers."""
import logging
import time

import kubernetes

KUBERNETES_EVENT_TYPE_MODIFIED = "MODIFIED"

KNATIVE_RESOURCE_GROUP = "serving.knative.dev"
KNATIVE_SERVICE_PLURAL = "services"

WATCH_TIMEOUT = 600

logger = logging.getLogger(__package__)


def _handle_namespaced_service_is_ready_event(
    namespace: str, name: str, event: dict
) -> bool:
    """Handle the ready event for Knative services."""
    try:
        if event["type"] != KUBERNETES_EVENT_TYPE_MODIFIED:
            return False

        metadata = event["object"]["metadata"]
        if metadata["namespace"] != namespace or metadata["name"] != name:
            return False

        return all(
            condition["status"] == str(True)
            for condition in event["object"]["status"]["conditions"]
        )
    except KeyError:
        return False


class KnativeTimeoutError(Exception):
    """Raised when the creation or update of a Knative services reaches a timeout."""


class KnativeServingV1Api:
    """Knative Serving v1 API.

    All methods are namespaced, only top-level Service resource is implemented.

    Unlike APIs provided by the Kubernetes client for Python, it waits for service
    readiness by default.
    """

    RESOURCE_VERSION = "v1"

    def __init__(self, client: kubernetes.client.ApiClient):
        self._custom_objects_api = kubernetes.client.CustomObjectsApi(client)
        self._watch = kubernetes.watch.Watch()

    def _get_knative_namespaced_service_params(self, namespace: str) -> dict[str, str]:
        return {
            "group": KNATIVE_RESOURCE_GROUP,
            "version": self.RESOURCE_VERSION,
            "plural": KNATIVE_SERVICE_PLURAL,
            "namespace": namespace,
        }

    def create_namespaced_service(self, namespace: str, body: dict) -> dict:
        """Create a Knative service."""
        params = self._get_knative_namespaced_service_params(namespace)
        service = self._custom_objects_api.create_namespaced_custom_object(
            **params, body=body
        )

        logger.info(
            "Waiting for service '%s' to be ready in namespace '%s'",
            service["metadata"]["name"],
            service["metadata"]["namespace"],
        )
        start = time.time()
        for event in self._watch.stream(
            self._custom_objects_api.list_namespaced_custom_object,
            timeout_seconds=WATCH_TIMEOUT,
            **params,
        ):
            if _handle_namespaced_service_is_ready_event(
                namespace=service["metadata"]["namespace"],
                name=service["metadata"]["name"],
                event=event,
            ):
                self._watch.stop()
                logger.info(
                    "Service '%s' ready in namespace '%s' after %ss",
                    service["metadata"]["name"],
                    service["metadata"]["namespace"],
                    int(time.time() - start),
                )
                return service

        raise KnativeTimeoutError(
            f"timeout during service '{service['metadata']['name']}' creation."
        )

    def update_namespaced_service(self, namespace: str, name: str, body: dict) -> dict:
        """Update a Knative service."""
        params = self._get_knative_namespaced_service_params(namespace)
        service = self.get_namespaced_service(namespace=namespace, name=name)
        updated_service = self._custom_objects_api.patch_namespaced_custom_object(
            **params, name=name, body=body
        )

        # Don't wait for the service to become ready if Knative generation is unchanged:
        if (
            updated_service["metadata"]["generation"]
            == service["metadata"]["generation"]
        ):
            return updated_service

        logger.info(
            "Waiting for service '%s' to be ready in namespace '%s'",
            updated_service["metadata"]["name"],
            updated_service["metadata"]["namespace"],
        )
        start = time.time()
        for event in self._watch.stream(
            self._custom_objects_api.list_namespaced_custom_object,
            timeout_seconds=WATCH_TIMEOUT,
            **params,
        ):
            if _handle_namespaced_service_is_ready_event(
                namespace=updated_service["metadata"]["namespace"],
                name=updated_service["metadata"]["name"],
                event=event,
            ):
                self._watch.stop()
                logger.info(
                    "Service '%s' ready in namespace '%s' after %ss",
                    updated_service["metadata"]["name"],
                    updated_service["metadata"]["namespace"],
                    int(time.time() - start),
                )
                return updated_service

        raise KnativeTimeoutError(
            f"timeout during service '{updated_service['metadata']['name']}' update."
        )

    def delete_namespaced_service(self, namespace: str, name: str) -> dict:
        """Delete a Knative service."""
        return self._custom_objects_api.delete_namespaced_custom_object(
            **self._get_knative_namespaced_service_params(namespace), name=name
        )

    def get_namespaced_service(self, namespace: str, name: str) -> dict:
        """Get a Knative service."""
        return self._custom_objects_api.get_namespaced_custom_object(
            **self._get_knative_namespaced_service_params(namespace), name=name
        )

    def list_namespaced_service(self, namespace: str) -> dict:
        """List Knative services."""
        return self._custom_objects_api.list_namespaced_custom_object(
            **self._get_knative_namespaced_service_params(namespace)
        )
